; ____________________________________________________________________
;|                                                                    |
;| This file is part of the project:                                  |
;|                                                                    |
;| ..::FreshLib::.. - portable, assembler library and GUI toolkit.    |
;|____________________________________________________________________|
;|                                                                    |
;|                          This file                                 |
;|                          _________                                 |
;|   Author:                                                          |
;|                                                                    |
;|   Title: StrLib.asm - OS independent string manipulation library.  |
;|                                                                    |
;|   OS:                                                              |
;|                                                                    |
;|   Notes and changes:                                               |
;|     Uses memory.asm library for memory functions.                  |
;|                                                                    |
;|                                                                    |
;|____________________________________________________________________|


STR_MINSTRLEN = 16      ; 16 bytes minimum string length

; NumToStr flags
ntsSigned = $00000
ntsUnsigned = $10000
ntsZeroTerminated = $20000
ntsFixedWidth     = $40000

ntsBin  = $0200
ntsQuad = $0400
ntsOct  = $0800
ntsDec  = $0a00
ntsHex  = $1000

struc TString {
  .capacity dd ?        ; length of allocated buffer.
  .len      dd ?        ; length of the string.
  label .data byte
}
virtual at -(sizeof.TString)
  TString TString
  sizeof.TString = $-TString
end virtual


struc string [value] {
common
local ..len, ..capacity, ..pad1, ..pad2
      virtual
        align 4
        ..pad1 = $ - $$
      end virtual
      db  ..pad1 dup 0

      dd  ..capacity
      dd  ..len
label . byte
forward
        db  value
common
  ..len = $ - .
      virtual
        align 4
        ..pad2 = $ - $$
      end virtual
      db  ..pad2 dup 0
  ..capacity = $ - .
      dd  0
}


; String format:
;   Let esi points to StrLib string. Then:
;   dword [esi-8] contains allocated memory size for the string.
;   dword [esi-4] contains length of the string in bytes.
;   [esi] .. [esi+length] contains the string itself.
;   at the end of the string, always have at least one dword zero, so it is safe to
;   scan the string by dwords.


; < Library functions >


;______________________________________________________________________________________
;
;  Creates new empty string and returns pointer to it.
;  Arguments:
;    NONE
;  RETURNS:
;    CF=0 -> no error; ebx = pointer to the new created string.
;    CF=1 -> memory allocation error. ebx=0
;______________________________________________________________________________________
proc StrNew
begin
        stdcall GetMem, STR_MINSTRLEN+sizeof.TString+4
        jc      .finish

        lea     eax, [eax+sizeof.TString]
        mov     [eax+TString.capacity], STR_MINSTRLEN

.finish:
        return
endp


;__________________________________________________________________________
;
; Delete string.
; Arguments:
;   .pString - string created with StrNew, or other StrLib procedure.
; Returns:
;   Nothing
;__________________________________________________________________________
proc StrDel, .pString
begin
        push    [.pString]
        sub     dword [esp], sizeof.TString
        call    FreeMem
        return
endp



;__________________________________________________________________________
;
; Duplicates given string, and returns a pointer to new one
; Arguments:
;   .pString - pointer to the string, allocated with StrLib
; Returns:
;   CF=0; eax - pointer to the duplicated string.
;   CF=1; eax - NULL
;__________________________________________________________________________
proc StrDup, .pString
begin
        stdcall StrNew
        jc      .exit

        stdcall StrCopy, eax, [.pString]
        jnc     .exit

        stdcall StrDel, eax
        xor     eax, eax
        stc
.exit:
        return
endp



proc StrDupAsciiZ, .pAsciiZ
begin
        stdcall StrNew
        jc      .exit

        stdcall StrCopyAsciiZ, eax, [.pAsciiZ]
        jnc     .exit

        stdcall StrDel, eax
        xor     eax, eax
        stc
.exit:
        return
endp



;__________________________________________________________________________
;
; proc StrSearchLen, .pString
; Searches to the end of ASCIIZ string and returns the length of the string.
;
; Arguments:
;  pString - pointer to the string.
; Returns:
;  ecx - length of the string in bytes.
;__________________________________________________________________________
proc StrSearchLen, .pString
begin
        mov     ecx, [.pString]
.scan:
        cmp     byte [ecx], 0
        lea     ecx, [ecx+1]
        jne     .scan

        stc
        sbb     ecx, [.pString]
        return
endp


;__________________________________________________________________________
; proc StrSearchLenFast, .pString
; The same as StrSearchLen, but works on dwords and is much faster.
; The string must ends with at least one aligned dword of NULL.
; This is the case with every string from StrLib, but not always with
; system provided strings.
;
; Arguments:
;  pString - pointer to the string.
; Returns:
;  ecx - length of the string in bytes.
;__________________________________________________________________________
proc StrSearchLenFast, .pString
begin
        mov     ecx, [.pString]
.scan:
        cmp     dword [ecx], 0
        lea     ecx, [ecx+4]
        jne     .scan

        lea     ecx, [ecx-4]

; so search back
.byteloop:
        lea     ecx, [ecx-1]
        cmp     byte [ecx], 0
        je      .byteloop

        lea     ecx, [ecx+1]
        sub     ecx, [.pString]
        return
endp





;___________________________________________________________________________
;
; Ensures that the string can take [.capacity] bytes.
; If the function resize the string, the new size is multiple of 4.
;
; Arguments:
;   .pString - string pointer.
;   .capacity - desired string capacity.
; Returns:
;   eax - pointer to the string (can be different than passed in .pString)
;   CF=1 if error allocation. In this case eax still contains pointer, but
;   the capacity is not changed.
;___________________________________________________________________________
proc StrSetCapacity, .pString, .capacity
begin
        push    ecx

        mov     ecx, [.capacity]

        add     ecx, 3 + sizeof.TString + 4
        and     cl, $fc
        cmp     ecx, STR_MINSTRLEN + sizeof.TString + 4
        jae     @f
        mov     ecx, STR_MINSTRLEN + sizeof.TString + 4
@@:
        sub     [.pString], sizeof.TString

        stdcall ResizeMem, [.pString], ecx
        jc      .finish

.sizeok:
        lea     ecx, [ecx-(sizeof.TString+4)]
        mov     [eax], ecx

.finish:
        lea     eax, [eax+sizeof.TString]
        pop     ecx
        return
endp




;_______________________________________________________________________________________
;  Copies source to destination string.
;  Arguments:
;     source - source string. for StrCopyAsciiZ - can be any zero terminated string.
;              for StrCopy, must be strlib string.
;     dest -  destination string. Always strlib string.
;  Returns:
;     eax - pointer to destination string.
;     CF=1 -> error with memory allocation. source is not copied and destination is
;             not changed.
;_______________________________________________________________________________________
proc StrCopyAsciiZ, .dest, .source
begin
        push    ecx esi edi

        mov     esi, [.source]
        stdcall StrSearchLen, esi
        jmp     StrCopy.docopy
virtual
  call StrCopy
end virtual
endp



proc StrCopy, .dest, .source
begin
        push    ecx esi edi

        mov     esi, [.source]
        mov     ecx, [esi+TString.len]

.docopy:
        stdcall StrSetCapacity, [.dest], ecx
        jc      .finish

        mov     edi, eax
        mov     [eax+TString.len], ecx
        push    ecx
        shr     ecx, 2
        rep movsd
        pop     ecx
        and     ecx, 3
        rep movsb

        mov     dword [edi], ecx
        add     edi, 3
        and     edi, $fffffffc
        mov     dword [edi], ecx
        clc

.finish:
        pop     edi esi ecx
        return
endp




proc StrCopyPart, .dest, .source, .pos
begin
        push    ecx esi edi

        mov     eax, [.dest]
        mov     esi, [.source]
        mov     ecx, [esi+TString.len]
        cmp     [.pos], ecx
        jae     .finish

        sub     ecx, [.pos]
        add     esi, [.pos]

        stdcall StrSetCapacity, [.dest], ecx
        jc      .finish

        mov     edi, eax
        mov     [eax+TString.len], ecx

        push    ecx
        shr     ecx, 2
        rep movsd
        pop     ecx
        and     ecx, 3
        rep movsb

        mov     dword [edi], ecx
        add     edi, 3
        and     edi, $fffffffc
        mov     dword [edi], ecx
        clc

.finish:
        pop     edi esi ecx
        return
endp


;__________________________________________________________________________________
; Splits the string on two strings, at position [.pos]
; Arguments:
;   .pString - pointer to string to be splitted.
;   .pos     - position where to split the string.
; Returns:
;   eax - pointer to the new created string with second part of the string.
;         the original string does not reallocate memory and it's capacity
;         and the pointer will remains the same.
;__________________________________________________________________________________

proc StrSplit, .pString, .pos
begin
        push    ecx

        stdcall StrNew

        mov     ecx, [.pString]
        mov     ecx, [ecx+TString.len]
        cmp     [.pos], ecx
        jae     .ready

        stdcall StrCopyPart, eax, [.pString], [.pos]

        mov     ecx, [.pString]
        push    [.pos]
        pop     [ecx+TString.len]  ; new length of the string.

        add     ecx, [ecx+TString.len]

        mov     dword [ecx], 0
        add     ecx, 3
        and     cl, $fc
        mov     dword [ecx], 0

.ready:
        pop     ecx
        return
endp








;_______________________________________________________________________________________
; Compares two strings - case sensitive.
; Returns CARRY = 1 if the strings are equal.
; Returns CARRY = 0 if the strings are different.
;_______________________________________________________________________________________
proc StrCompCase, .str1, .str2
begin
        push    eax ecx esi edi

        mov     esi, [.str1]
        mov     edi, [.str2]

        cmp     esi, 0
        je      .noteq

        cmp     edi, 0
        je      .noteq

        mov     ecx, [esi+TString.len]
        cmp     ecx, [edi+TString.len]
        jne     .noteq

        repe cmpsb
        jne     .noteq

        stc
        pop     edi esi ecx eax
        return

.noteq:
        clc
        pop     edi esi ecx eax
        return
endp


;_______________________________________________________________________________________
; Compares two strings - case NOT sensitive.
; Returns CARRY = 1 if the strings are equal.
; Returns CARRY = 0 if the strings are different.
;_______________________________________________________________________________________
proc StrCompNoCase, .str1, .str2
begin
        push    eax esi edi
        stdcall StrPtr, [.str1]
        mov     esi,eax
        stdcall StrPtr, [.str2]
        mov     edi,eax

.cmploop:
        mov     al, [esi]
        lea     esi, [esi+1]
        mov     ah, [edi]
        lea     edi, [edi+1]

        test    al,al
        jz      .eos
        test    ah,ah
        jz      .noteq

        or      eax, $2020
        cmp     al, ah
        je      .cmploop

.noteq:
        clc
        pop     edi esi eax
        return

.eos:
        test    ah,ah
        jnz     .noteq

        stc
        pop     edi esi eax
        return
endp





;__________________________________________________________________________________
; StrCat appends one string to another
; Arguments:
;   dest - destination string (handle only)
;   source - source string
; Returs:
;   eax - pointer to destination string.
;   CF - error flag.
;__________________________________________________________________________________
proc StrCat, .dest, .source
begin
        push    ecx esi edi

        mov     eax, [.dest]
        mov     esi, [.source]
        mov     edi, eax

        mov     ecx, [esi+TString.len]
        add     ecx, [edi+TString.len]
        cmp     ecx, [edi+TString.capacity]
        jbe     .sizeok

        stdcall StrSetCapacity, edi, ecx
        jc      .finish
        mov     edi, eax

.sizeok:
        xchg    [edi+TString.len], ecx
        add     edi, ecx

        mov     ecx, [esi+TString.len]
        push    ecx
        shr     ecx, 2

        rep movsd
        pop     ecx
        and     ecx, 3
        rep movsb

        mov     dword [edi], 0

        clc

.finish:
        pop     edi esi ecx
        return
endp


;__________________________________________________________________________________
; StrCharPos returns a pointer to the first occurence of a given char
;   in specified string
; Arguments:
;   Char - char to look for
;   hString -  string to search
; Returns:
;   ebx - offset of the char in the string.
;   CF=1 if the char is found.
;   CF=0 if the char is not found. EBX is not changed in this case.
;__________________________________________________________________________________
proc StrPosChar, .pString, .char
begin
        push    eax ecx ebx

        mov     ebx, [.pString]
        mov     ecx, [esi+TString.len]

        mov     al, byte [.char]

.search:
        cmp     [ebx], al
        je      .found
        lea     ebx, [ebx+1]
        dec     ecx
        jnz     .search

        clc
        pop     ebx ecx eax
        return

.found:
        sub     ebx, [.pString]
        lea     esp, [esp+4]
        stc
        pop     ecx eax
        return
endp


;__________________________________________________________________________________
; StrPos returns the offset of the first occurence of a pattern string
;   in another string
; Arguments:
;   hPattern - 'pattern' string
;   hString -  string to search
; Returns: a pointer to the pattern string in source , or NULL if pattern string
; doesn't occur in the string to search
;__________________________________________________________________________________
proc StrPos, .pString, .pPatern
begin
        push    esi edi eax ecx ebx

        mov     esi, [.pString]
        mov     edi, [.pPatern]

        mov     ecx, [esi+TString.len]
        mov     ebx, [edi+TString.len]
        test    ebx, ebx
        jz      .found

        cmp     ebx, ecx
        ja      .notfound       ; the parern is longer than string.

        sub     ecx, ebx
        dec     esi
        lea     ecx, [ecx+2]

.outer:
        dec     ecx
        jz      .notfound

        mov     ebx, [edi+TString.len]
        lea     esi, [esi+1]

.inner:
        mov     al, [esi+ebx-1]
        cmp     al, [edi+ebx-1]
        jne     .outer

        dec     ebx
        jnz     .inner

.found:
        sub     esi, [.pString]
        mov     [esp], esi
        stc
        pop     ebx ecx eax edi esi
        return

.notfound:
        clc
        pop     ebx ecx eax edi esi
        return
endp




;__________________________________________________________________________________
; StrInsert inserts one string into another at specified pos
; Arguments:
;   dest - destination where the source will be inserted.
;   source -  string to insert
;   pos    - where to insert.
; Returns:
;   eax - pointer to destination string.
;__________________________________________________________________________________
proc StrInsert, .dest, .source, .pos
begin
        stdcall StrSplit, [.dest], [.pos]
        push    eax eax
        stdcall StrCat, [.dest], [.source]
        stdcall StrCat, eax ; source from the stack.
        stdcall StrDel; from the stack.
        return
endp



;__________________________________________________________________________________
; Converts strings to Lower Case
; First parameter = String to Convert to lower case
; corrupts eax
;__________________________________________________________________________________
proc StrLCase, .ptrString
begin
        push    eax edi
        mov     edi, [.ptrString]
        mov     eax, [edi+TString.len]
        dec     edi
.loop:
        inc     edi
        dec     eax
        js      .endstring

        cmp     byte [edi], 'A'
        jb      .loop
        cmp     byte [edi], 'Z'
        ja      .loop

        or      byte [edi], 20h
        jmp     .loop

.endstring:
        pop     edi eax
        return
endp


;__________________________________________________________________________________
; Converts strings to Upper Case
; First parameter = String to Convert to upper case
; corrupts eax
;__________________________________________________________________________________
proc StrUCase, .ptrString
begin
        push    eax edi
        mov     edi, [.ptrString]
        mov     eax, [edi+TString.len]
        dec     edi
.loop:
        inc     edi
        dec     eax
        js      .endstring

        cmp     byte [edi], 'a'
        jb      .loop
        cmp     byte [edi], 'z'
        ja      .loop

        and     byte [edi], ~20h
        jmp     .loop

.endstring:
        pop     edi eax
        return
endp




;__________________________________________________________________________________
; _NumToStr converts the number in eax to the string in any radix approx. [2..26]
; Arguments:
;   edi - pointer to the string buffer
;   ecx - radix
;   eax - number to convert.
; There is no parameter check, so be careful.
; returns: edi points to the end of a converted number
;__________________________________________________________________________________
proc _NumToStr
begin
    test  eax, eax
    jns   _NumToStrU
    neg   eax
    mov   byte [edi],"-"
    inc   edi
endp


proc _NumToStrU
begin
    cmp   eax,ecx
    jb    .lessA

    xor   edx,edx
    div   ecx
    push  edx
    call  _NumToStrU
    pop   eax

.lessA:
    cmp   al, 10
    sbb   al, 69h
    das
    stosb
    return
endp

;_____________________________________________________
; NumToStrF:
;   Converts signed integer value to string.
; NumToStrUF:
;   Converts unsigned integer value to string.
;
; edi - pointer to string buffer
; eax - Number to convert
; ecx - radix from 2 to 26
; esi - length of the number in chars
;
; returns: edi - pointer to the end of converted num
;
; Note: Don't use 1 as radix.
;_____________________________________________________
proc NumToStrF
begin
        test    eax,eax
        jns     NumToStrUF
        neg     eax
        mov     byte [edi],'-'
        push    esi
        dec     esi
        add     edi,esi
        push    edi
        jmp     NumToStrUF.loopc
endp

proc NumToStrUF
        begin
        push    esi
        add     edi, esi
        push    edi
        dec     edi
.loopc:
        xor     edx,edx
        div     ecx
        xchg    al,dl
        cmp     al,$0a
        sbb     al,$69
        das
        mov     [edi],al
        dec     edi
        xchg    al,dl
        dec     esi
        jnz     .loopc
        pop     edi
        pop     esi
        return
endp


;___________________________________________________________
; NumToStr - converts number to any radix.
; num - number to convert
; str - handle of the string. If NULL - creates new string.
; index - Offset in string where to put converted number.
; flags:
;   byte 0 - number of digits if ntsFixedWidth is set.
;   byte 1 - contains radix for the convertion.
;   byte 2,3 - flags.
; Returns:
;   eax - handle of the string (new one or passed in [str])
;   edx - pointer to the string.
;___________________________________________________________
proc NumToStr, .num, .strng, .index, .flags
.max_len dd ?
.len_tab rb 20      ; table that holds maximum number of digits in given system
begin
        push    ebx ecx edx esi edi

        mov     [.len_tab+2],32+2       ; binary number - max. 32 digits (+2 bytes for '-' character and NULL terminator)
        mov     [.len_tab+4],16+2       ; quad number - max. 16 digits
        mov     [.len_tab+8],11+2       ; octal number - max. 11 digits
        mov     [.len_tab+10],9+2       ; decimal number - max. 9 digits
        mov     [.len_tab+16],8+2       ; hexadecimal number - max. 8 digits
        movzx   ebx,byte [.flags+1]     ; load radix into ebx
        movzx   eax,byte[.len_tab+ebx]
        mov     [.max_len],eax          ; store max. number of digits
        mov     eax,[.strng]
        or      eax,eax
        jnz     .string_exists
        stdcall StrNew
        mov     [.strng],eax

  .string_exists:
        test    eax,0xffff0000
        jnz     @f                      ; don't resize if [str] isn't a handle
        mov     ebx,eax
        mov     eax,[.max_len]
        add     eax,[.index]
        mov     edx,eax
        stdcall StrSearchLen, ebx
        cmp     eax,edx
        ja      @f                      ; don't resize string if it has enough place for converted number

        stdcall StrSetLength,ebx, edx

  @@:   ; determine which conversion func to use
        mov     eax,[.flags]
        mov     edx,eax
        and     eax,ntsFixedWidth
        jnz     .fixed_width
        mov     eax,edx
        and     eax,ntsUnsigned
        jnz     .unsigned
        mov     ebx,_NumToStr
        jmp     .got_func
  .unsigned:
        mov     ebx,_NumToStrU
        jmp     .got_func

  .fixed_width:
        movzx   esi, byte [.flags]         ; load fixed width into esi
        mov     eax, edx
        and     eax, ntsUnsigned
        jnz     .fixed_unsigned
        mov     ebx,NumToStrF
        jmp     .got_func
  .fixed_unsigned:
        mov     ebx,NumToStrUF
  .got_func:
        stdcall StrPtr, [.strng]
        add     eax,[.index]
        mov     edi,eax
        movzx   ecx,byte [.flags+1]       ; load radix into ecx
        mov     eax, [.num]
        call    ebx                     ; call low-level convertion routine
        mov     eax, [.flags]
        and     eax, ntsZeroTerminated
        jz      .ret
        mov     byte [edi], 0

  .ret:
        pop     edi esi edx ecx ebx
        mov     eax, [.strng]
        return
endp



;-------------------------------------------------------
; function StrToNum
;   Converts specified string into a number
;
; Arguments:
;   hString - handle/pointer of the string containing
;     number to convert. It doesn't have to be ended by
;     NULL, any other character will stop conversion.
;     Number to convert must be decimal.
;
; Return:
;   eax - converted number
;
; Note: in case of failture (first char of given pointer
;   isn't a number) function returns -1.
;-------------------------------------------------------
proc StrToNum, .hString
begin
        push    ebx edx esi
        xor     ebx,ebx         ; ebx will store our number
        stdcall StrPtr, [.hString]
        mov     esi,eax
        xor     eax,eax
        mov     al,[esi]
        cmp     al,'0'
        jb      .error
        cmp     al,'9'
        jbe     .digit
        jmp     .error
     .digit:
        sub     al,'0'
        add     ebx,eax
        inc     esi
        mov     al,[esi]
        cmp     al,'0'
        jb      .finish
        cmp     al,'9'
        ja      .finish
        mov     edx,ebx         ; multiply ebx by 10
        shl     ebx,3
        add     ebx,edx
        add     ebx,edx
        jmp     .digit
     .finish:
        mov     eax,ebx
        pop     esi edx ebx
        return

     .error:
        xor     eax,eax
        dec     eax
        pop     esi edx ebx
        return
endp


;-------------------------------------------------------
; function StrCharCat
;   Addes up to 4 chars to the end of a string.
;
; Arguments:
;   ptrString - string to append
;   char - char(s) to add
; Returns
;   eax: pointer to the string.
;-------------------------------------------------------
proc StrCatChar, .ptrString, .char
begin
        push    ebx ecx

        mov     eax, [.ptrString]
        mov     ecx, [eax+TString.len]
        lea     ebx, [ecx+4]
        cmp     ebx, [eax+TString.capacity]
        jbe     .sizeok

        stdcall StrSetCapacity, eax, ebx

        pushd   [.char]
        popd    [eax+ecx]
        xor     ebx, ebx
        mov     [eax+ecx+4], ebx
        dec     ecx
.len:
        inc     ecx
        cmp     [eax+ecx], bl
        jne     .len

        mov     [eax+TString.len], ecx
        pop     ecx ebx
        return
endp